#!/usr/bin/perl

#
# Copyright (C) 2005 Schneelocke
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 2 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc., 59
# Temple Place, Suite 330, Boston, MA 02111-1307 USA
#

# ========== NO NEED TO CONFIGURE ANYTHING BELOW THIS LINE ==========

# faster, and allows for easier GF->NW conversion
use integer;
use Getopt::Long qw(:config pass_through);

our $VERSION = '0.02';

# helper routine: stringify a hash (in a format suitable for our purposes).
sub dump_hash {
	my %hash = @_;
	my $retval = "";

	foreach (sort keys %hash) {
		$retval .= $_ . $hash{$_} . " ";
	}

	return $retval;
}

### MAIN PROGRAM STARTS HERE ###

$showhelp = 0;

$result = GetOptions(
		"help"		=> \$showhelp,
	);

if($showhelp) {
	print "tasdoom_unscramble.pl $VERSION\nCopyright (C) 2005 Schneelocke, licensed under the terms of the GNU GPL v2\n(or, at your option, any later version)\n\n";
	print "Usage: cat tasdoom_demo.txt | tasdoom_unscramble.pl >unscrambled_demo.txt\n\n";
	print "This program attempts to unscramble a demo recorded with TASDOOM [1] so it can\n";
	print "be played back with doom2.exe, PrBoom etc. In order to unscramble a demo, you\n";
	print "should:\n\n";
	print "1. Convert the demo to text format using Uwe Girlich's LMPC [2];\n";
	print "2. Unscramble the demo using this program;\n";
	print "3. Convert the demo back to binary format using LMPC again.\n\n";
	print "If you get any warnings about unknown commands etc., please send me the demo\n";
	print "you're trying to unscramble - my email address is tasdoom@gl00on.net.\n\n";
	print "1. http://www.3ddownloads.com/doomworld/tas/tasdoom.zip\n";
	print "2. http://www.planetquake.com/demospecs/lmpc-stable/\n\n";
	print "This program is free software; you can redistribute it and/or\n";
	print "modify it under the terms of the GNU General Public License\n";
	print "as published by the Free Software Foundation; either version 2\n";
	print "of the License, or (at your option) any later version.\n\n";
	print "This program is distributed in the hope that it will be useful,\n";
	print "but WITHOUT ANY WARRANTY; without even the implied warranty of\n";
	print "MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n";
	print "GNU General Public License for more details.\n\n";
	print "You should have received a copy of the GNU General Public License\n";
	print "along with this program; if not, write to the Free Software\n";
	print "Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.\n";
	exit(0);
}


print STDERR "tasdoom_unscramble.pl $VERSION\n";
while(<>) {

	($in_data_section = 1, print("DataStart\n"), next) if(m/^DataStart/);
	($in_data_section = 0, print("DataEnd\n"), next) if(m/^DataEnd/);

	# if we're not in the data section, don't parse anything and instead just print it verbatim.
	unless($in_data_section) {
		s/1\.10/1\.9/ if(m/Version:/);
		print;
		next;
	}

	chomp;
	s/(\s*#.*$)//;

	# read commands into a hash (%commands).
	%commands = ();
	foreach $raw_command (split /\s+/, $_) {

		# split into command (e.g., "GF") and value (e.g., 50).
		if($raw_command =~ m/^([A-Z]+)(0x[0-9a-f]{2}|\d*$)/) {;

			($command, $value) = ($1, $2);

			# take care of hexadecimal values for "U0xff" etc.
			if($value =~ m/^0x/) {
				$value = oct $value;
			}

			$commands{$command} = $value;

		} else {

			warn "Cannot parse command: $raw_command\n";

		}
	}

	# The output is collected in this string.
	$result = "";

	# GB
	# (not handled yet)

	# GF
	if(defined $commands{GF}) {

		if($commands{GF} & 1) {
			$result .= "FW ";
		}

		if($commands{GF} & 2) {
			$result .= "UT ";
		}
		
		if($commands{GF} > 3) {

			$commands{GF} += 4;
			$commands{GF} /= 8;

			$result .= "NW" . $commands{GF} . " ";

		}
		
		delete $commands{GF};
	}

	# NW
	if(defined $commands{NW}) {

		if(defined $commands{UT}) {

			if(defined $commands{FW}) {

				$result .= "TL" . (8 * $commands{NW} - 1) . " ";
				delete $commands{FW};

			} 
			else {
				$result .= "TL" . (8 * $commands{NW} - 2) . " ";
			}

			delete $commands{UT};

		} 
		else {

			if(defined $commands{FW}) {

				$result .= "TL" . (8 * $commands{NW} - 3) . " ";
				delete $commands{FW};

			}
			else {
				$result .= "TL" . (8 * $commands{NW} - 4) . " ";
			}
		}

		delete $commands{NW};
	}


	# SG
	if(defined $commands{SG}) {

		$result .= "TR" . ((7 - $commands{SG}) * 4 + 2) . " ";
		delete $commands{SG};

	}

	# SL
	if(defined $commands{SL}) {

		$result .= "GB" . $commands{SL} . " ";
		delete $commands{SL};

	}

	# SR
	if(defined $commands{SR}) {

		$result .= "GF" . $commands{SR} . " ";
		delete $commands{SR};

	}

	# TL
	if(defined $commands{TL}) {

		$result .= "SR" . $commands{TL} . " ";
		delete $commands{TL};

	}

	# TR
	if(defined $commands{TR}) {

		$result .= "SL" . $commands{TR} . " ";
		delete $commands{TR};

	}

	# U
	if(defined $commands{U}) {

		if($commands{U} < 128) {

			$result .= "TL" . $commands{U} . " ";


		}
		else {
			$result .= "TR" . (256 - $commands{U}) . " ";
		}

		delete $commands{U};

		# there can also be an FW or UT (or both) here, which doesn't seem to have any effect.
		delete $commands{FW};
		delete $commands{UT};

	}

	### NOTE: UT and FW have to come after the above.

	# FW
	if(defined $commands{FW}) {

		if(defined $commands{UT}) {

			$result .= "TL3 ";
			delete $commands{UT};

		} 
		else {
			$result .= "TL1 ";
		}

		delete $commands{FW};
	}

	# UT
	if(defined $commands{UT}) {
		
		$result .= "TL2 ";
		delete $commands{UT};

	}

	# WT
	if(defined $commands{WT}) {

		$result .= "WT ";
		delete $commands{WT};

	}

	if($result ne "") {
		print "$result\n";
	} 
	else {
		warn "No valid commands found!\n";
	}

	if($c = scalar keys %commands) {
		warn "$c uninterpreted command" . ($c == 1 ? "" : "s") . " found: ", dump_hash(%commands), "; input line was $_\n";
	}

}

